/*
            CSE687 Project 1 "Stack information cluster in C and C++"

Name: Gang Cheng      SSN: 223-61-3010

1. Description of the test program

The program reads a stream from the file specified in command line (if any), or 
by default from a file named "proj1.txt" in the working directory if no file specified
in the command line or it failed to open that file, into a stack named stack1.
It then copies all elements in stack1 into another stack(stack2). 
After it displays the content in stack1 in the order from top to bottom, it moves all
elements from stack1 into stack3. After all elements are displayed, stack3 is copied to
stack4. Then the content of stack2 and stack4 are displayed, respectively.

2. Stack structure

I am using link list as basic cell for the stack so that only the top pointer of a stack
is kept to trace a stack.  While each push operation dynamically allocated a cell from
operating system, each pop peration frees the cell popped up. The advantage is that the 
maximum size of required stacks is determined dynamically by runtime memory avaiable in 
the whole system. The drawback is the allocation and free operation on memory goes 
togather with the growing and decresing of stacks, which would decrease the performance 
of some applications using stacks. For our test program, there is no obvious slow-down 
in terms of performance. In addition, all the techniques and principles can easily be
applied to any other stack structures, such as linear list(array), 

3. Information cluster of the stack class

 . stack constructor ---- initialize a stack
 . stack destructor  ---- abandon a stack, free the memory it occupied 
 . push operation    ---- push an element into a stack
 . pop  operation    ---- pop an element from a stack (returns NULL if stack is empty,
                          so it can also be used as test-empty operation for a stack)
 . show operation    ---- from top to bottom, display on standard output all elements
                          in a stack
 . scan operation    ---- find first location(from top to bottom) on a stack for a 
	             specified element(returns NULL if no such element on the stack)
 . copy operation    ---- copy a stack to another one (make an identical structure)

4. C implementation

I am following the techniques used in X-Window Toolkit library which is the base 
for the well-known OSF/Motif widgets for X programming, to implement the Stack 
class in C. The basic idea is to use structs, macro definitions and typedefs to 
implement OOP. In order to implement concepts such as Objects and Data 
Abstraction, Class and Methods, Inheritance in OOP, following  general class 
structure is used:

       (Class data structure)
     Pointer to super class ---> (none in our stack class)
     Methods for this class ---> (Pointers to functions that operate on instance 
	   of this class, point to all information cluster functions in our stack class)
     Data common to all 
       instance of this class
     

         (instance 1)        ...                                    (instance 2)
     Pointor to Class data structure ---->(above structure)     (same structure as 1)
     Instance ID (stack id)
     Instance-specific 
       data (top pointor in our stack class)

The typical use of a class stack in this implementation is provided by routines below:

 1. Stack stack-id; (required by the language) // define a stack 
 2. Stack_Create(stack-id);                    // initialize a stack 
 3. Stack_Push(stack-id, element);             // push operation 
    Stack_Pop(stack-id);                       // pop operation 
    Stack_Show(stack-id);                      // show a stack 
    Stack_Scan(stack-id, element);             // locate an element on a stack 
    Stack_Stack(stack-id-from, stack-id-to);   // copy a stack to another stack
 4. Stack_Destroy(stack-id);                   // destroy a stack 

The user has to at least follow 1, 2 and 4 to use the stack information cluster.
By the above interface routines, C implementation allows a client program to
safely perform common stack operations such as push and pop without directly
accessing the stack internals. It provides an example of how to support data abstraction,
mutiple instance and other OOP concepts in C.
 
5. Comment

It is very easy to see, from the source codes of the implementations in C and C++,
the advantages provided by C++ to implement the OOP. 
This project shows the implementation of modular programming, data abstraction,
multiple objects(instance) and encapsulation (data hiding).

. modular programming: both versions support it by using seperately compiled file.

. data abstraction(DA): C++ uses public member functions of a class to support 
  the data abstraction, while C version uses a set of pre-defined macros to provide 
  interface routines to the user. Because the inherent mechnisim in C++, the 
  implementation of DA is very natural and easy to identify. In C version, 
  error-proned macro definitions and pointer manipulation has to be used to 
  achieve the DA, which leads to the difficult to maintain the program 
  (low extendability and reuseability) due to the inconsistancy of the use of 
  macros and the invoked function routines and may suffer a little  performance 
  due to the unnecessary pointer manipulation. 

. multiple objects: Being a data type like other built-in data type such as integer, 
  the class in C++ provides a easy and natural way to support multiple instances of 
  a (user-defined complex) data type. C version has to a lot of complicated pointer
  manipulation.

. data encapsulation: The private part of a class in C++ provides a truely data
  encapsulation, while the C version by no means prevents  the user from trying 
  to use the private data (such as stack top pointer in our project). 

. general comment: I was really amazed by the facilities provided in C++ to 
  support OOD in the language level, after finishing this simple but illustrating 
  project. At the first glimpse when I first learn to use C++, I was thinking
  it is totally redundant as I had claimed to myself that I could use the powerful 
  C to do OOD and C++ is merely a super shell(or library) of C. It proved that 
  I was wrong though this project as I found I spent more than double time 
  implementing C version but do the C++ version very easily and naturely.
*/

/***************************************************************
 * stack.h ---- header file for proj1.c and stack.c            *
 *                                                             *
 ***************************************************************/

#ifndef STACK_H

#define STACK_H

/* link list structure */
struct link_list {    
  char value;          
  struct link_list *next;
};

/* stack operation entries (Stack class structure) */
struct stack_common {  
  void (* push_op)(struct link_list **top,char element);    /* pointer to function push */
  char (* pop_op)(struct link_list **top);                  /* pointer to function pop  */
  void (* show_op)(struct link_list *top);                  /* pointer to function show */
  struct link_list (* scan_op)(struct link_list *top, char element);  /* pointer to function scan */
  void (* copy_op)(struct link_list *top_from, struct link_list **top_to); 
};

/* stack structure (instance structure) --- interface to define instance of the Stack class*/
typedef struct _Stack {   /*  Stack of integers */
  struct stack_common *stack_ops;  /* pointer to data structure of stack class (common part of the class) */
  struct link_list *stack_top;     /* instance-specific data (stack top pointor) */
} Stack;

/* function prototypes */
void push(struct link_list **top,char element);
char pop(struct link_list **top);
void show(struct link_list *top);          
struct link_list * scan(struct link_list *top, char element);
void copy(struct link_list *top_from, struct link_list **top_to); 

/* extern variables */
extern struct stack_common CLASS; 

/* macro definitions --- interface to do operations on the class instances(objects) */
#define Stack_Push(x,y) (*(((x).stack_ops)->push_op))(&((x).stack_top),y) /* Push y into stack x */
#define Stack_Pop(x) (*(((x).stack_ops)->pop_op))(&((x).stack_top))       /* Pop from stack x(return NULL if x is empty) */
#define Stack_Show(x) (*(((x).stack_ops)->show_op))((x).stack_top)        /* Show all elements in stack x from top to bottom */
#define Stack_Scan(x) (*(((x).stack_ops)->show_op))((x).stack_top,(y))    /* Query the first location from top of an element y 
									     on stack x(return NULL if y is not in x)  */
#define Stack_Copy(from,to) (*(((from).stack_ops)->copy_op))((from).stack_top,&((to).stack_top))  
#define Stack_Create(x) { (x).stack_ops = &CLASS; (x).stack_top = NULL;}  /* Create stack x  */
#define Stack_Destroy(x) { while (Stack_Pop(x)); (x).stack_ops = NULL; }  /* destroy stack x (free space if any) */

#endif STACK_H

/***************************************************************
 * stack.c ---- modul of information cluster for stack class   *
 *              (C version)                                    *
 ***************************************************************/

#include <stdio.h> 
/* #include "stack.h" */

/* the sole global variable to define a common structure for Stack class */
struct stack_common CLASS = { (void (*)(struct link_list **top,char element))(push),
			      (char (*)(struct link_list **top))(pop),
			      (void (*)(struct link_list *top)) (show),
			      (struct link_list (*)(struct link_list *top, char element))(scan),
			      (void (*)(struct link_list *top_from,struct link_list **top_to))(copy)};

/* push operation */
void push(struct link_list **top,char element)
{
  struct link_list *tmp;

  tmp = (struct link_list *) malloc(sizeof(struct link_list));

  tmp->value = element;
  tmp->next = *top;
  *top = tmp;

  return;
}

/* pop operation */
char pop(struct link_list **top)
{
  char ch;
  struct link_list *tmp = *top;

  if (*top == NULL) return NULL;

  ch = (*top)->value; 
  *top = (*top)->next;
  free(tmp);

  return ch;
}

/* display on stdout all elements in a stack (in the order from top to bottom) */
void show(struct link_list *top)
{
  char ch;
  struct link_list *tmp = top;  

  while (tmp) {
    putchar(ch = tmp->value);
    tmp = tmp->next;
  };

  return;
}

/* find location on a stack for a specified element */
struct link_list * scan(struct link_list *top, char element)
{
  struct link_list *tmp = top;
  
  while (tmp && (tmp->value != element)) tmp = tmp->next;
  
  return (tmp);
}

/* copy all elements in a stack to another stack */
void copy(struct link_list *top_from, struct link_list **top_to)
{
  struct link_list *tmp_from = top_from, *tmp_to, *tmp;

  if (top_from == *top_to) return; /* return if copy to itself */
  while ((tmp = *top_to) != NULL) {        /* free any space left */
    *top_to = (*top_to)->next;
    free(tmp);
  };

  if (top_from == NULL) 
    return;
  else {
    tmp_to = *top_to = tmp = (struct link_list *) malloc(sizeof(struct link_list));
    tmp_to->value = top_from->value;
    tmp_from = tmp_from->next;
  };

  /* copy elementwise from top to bottom */
  while (tmp_from) {
    tmp->next = tmp_to = (struct link_list *) malloc(sizeof(struct link_list));
    (tmp = tmp_to)->value = tmp_from->value;
    tmp_from = tmp_from->next;
  };
  tmp->next = NULL;

  return;
}
/***************************************************************
 * proj1.c ----  main program to test the information cluster  *
 *               for stack class (C version)                   *
 ***************************************************************/

/* #include <stdio.h>
#include "stack.h" */

#define default_file "proj1.txt" /* default file name used for input */

/* function prototypes */
void read_into_stack(FILE *fp, Stack *stack_id);
void write_onto_stack(Stack *stack_id1,Stack *stack_id2);
void error(char *s1, char *s2);

/* ---- <void error(char *s1, char *s2)> ---- */
/* error handler */
void error(char *s1, char *s2)
{
  printf("%s %s\n",s1,s2);
  exit(1);
}

/* <main program> */
/* test the information cluster for stack class
   tested member functions of class Stack:
   . stack constructor/destructor . stack copier
   . show all elements in a stack (from top to bottom) */
main(argc, argv) 
int argc; char *argv[];
{
  FILE *fp = NULL;
  Stack stack1,stack2,stack3,stack4; /* definition for stacks */

  /* open file specified on command line or the defaulted one and check i/o status */
  if (argc >1) {   /* open file specified on command line */
    if (!(fp = fopen(argv[1],"r"))) {
      printf("cannot open file: %s\n",argv[1]);
      printf("now try to open the default file: %s\n",default_file);
    };
  };
  if (!fp && (fp = fopen(default_file,"r")) == 0)   /* open default file */
    error("cannot open file %s\n",default_file);    
  
  printf("This is file %s\nProcessed by %s\n",
	 ((argc == 1) ? default_file : argv[1]), argv[0]);
  
  /* Create stack1 and stack3 */
  Stack_Create(stack1);
  Stack_Create(stack3);
  
  /* read the stream in file into stack1 */
  read_into_stack(fp,&stack1);
  
  /* copy stack1 to stack 3 */
  Stack_Copy(stack1,stack3);

  /* print a centered line */
  printf("\n--------------------- Reversed Stream -----------------------\n");

  /* show reversed stream in stack1 */
  Stack_Show(stack1);

  /* Create stack2 and stack4 */
  Stack_Create(stack2);
  Stack_Create(stack4);

  /* write the reversed stream from stack1 onto stack2 */
  write_onto_stack(&stack1,&stack2); 

  /* print a centered line */
  printf("\n--------------------- Normal Stream -------------------------\n");

  /* show  stream in stack2 */
  Stack_Show(stack2);  

  printf("\n"); /* end newline */

  /* copy stack2 to stack 4 */
  Stack_Copy(stack2,stack4);

  /* show once more using copied stacks */
  printf("\n------------------ Again Reversed Stream --------------------\n");
  Stack_Show(stack3);
  printf("\n------------------ Again Normal Stream ----------------------\n");
  Stack_Show(stack4);
  printf("\n"); 
  
  /* distroy stacks */
  Stack_Destroy(stack2);
  Stack_Destroy(stack1);
  Stack_Destroy(stack4);
  Stack_Destroy(stack3);

  exit (0);
}

/* ---- <void read_into_stack(FILE *fp, Stack *stack_id)> ---- */
/* read stream from a file into stack */
void read_into_stack(FILE *fp, Stack *stack_id)
{
  char ch;

  while ((ch = getc(fp)) != EOF) Stack_Push(*stack_id,ch);
  
  return;
}

/* ---- <void write_onto_stack(Stack *stack_id1, Stack *stack_id2> ---- */
/* move elements from a stack into another stack */
void write_onto_stack(Stack *stack_id1, Stack *stack_id2)
{
  char ch;

  while ((ch = Stack_Pop(*stack_id1)) != NULL) Stack_Push(*stack_id2,ch); 

  return;
}
